import os
from typing import Union
from importlib.metadata import version
from datetime import datetime, timezone


class Configuration:
    def __init__(self, path: str):
        self.path: str = path
        # config file specific
        self.src: str = None
        self.dst: str = None
        self.plt: str = None
        self.url: str = None
        self.static_url: str = None
        self.default_image_url: str = None
        self.title: str = None
        self.dformat: str = None
        self.l_dformat: str = None
        self.lsep_dformat: str = None
        self.force: bool = None

        # other
        self.version: str = version('pyssg')
        self.dformat_rss: str = '%a, %d %b %Y %H:%M:%S GMT'
        self.dformat_sitemap: str = '%Y-%m-%d'
        self.run_date_rss = datetime.now(tz=timezone.utc).strftime(self.dformat_rss)
        self.run_date_sitemap = \
        datetime.now(tz=timezone.utc).strftime(self.dformat_sitemap)


    def read(self):
        try:
            lines: list[str] = None
            with open(self.path, 'r') as f:
                lines = f.readlines()

            opts: dict[str, Union[str, bool]] = dict()
            for l in lines:
                kv: list[str] = l.split('=', 1)
                if len(kv) != 2:
                    raise Exception('wrong config syntax')

                k: str = kv[0].strip().lower()
                v_temp: str = kv[1].strip()
                # check if value should be a boolean true
                v: Union[str, bool] = v_temp\
                    if v_temp.lower() not in ['true', '1', 'yes']\
                    else True

                opts[k] = v

            try:
                self.src = opts['src']
            except KeyError: pass

            try:
                self.dst = opts['dst']
            except KeyError: pass

            try:
                self.plt = opts['plt']
            except KeyError: pass

            try:
                self.url = opts['url']
            except KeyError: pass

            try:
                self.static_url = opts['static_url']
            except KeyError: pass

            try:
                self.default_image_url = opts['default_image_url']
            except KeyError: pass

            try:
                self.title = opts['title']
            except KeyError: pass

            try:
                self.dformat = opts['date_formaT']
            except KeyError: pass

            try:
                self.l_dformat = opts['list_date_FORMAT']
            except KeyError: pass

            try:
                self.lsep_dformat = opts['list_sep_dATE_FORMAT']
            except KeyError: pass

            try:
                # if the parser above didn't read a boolean true, then take it
                # as a false anyways
                self.force = opts['force'] if opts['force'] is True else False
            except KeyError: pass

        except OSError: pass


    def fill_missing(self, opts: dict[str, Union[str, bool]]) -> None:
        if self.src is None:
            self.src = opts['src']

        if self.dst is None:
            self.dst = opts['dst']

        if self.plt is None:
            self.plt = opts['plt']

        if self.url is None:
            self.url = opts['url']

        if self.static_url is None:
            self.static_url = opts['static_url']

        if self.default_image_url is None:
            self.default_image_url = opts['default_image_url']

        if self.title is None:
            self.title = opts['title']

        if self.dformat is None:
            self.dformat = opts['date_format']

        if self.l_dformat is None:
            self.l_dformat = opts['list_date_format']

        if self.lsep_dformat is None:
            self.lsep_dformat = opts['list_sep_date_format']

        if self.force is None:
            self.force = opts['force']



import os
import shutil
from importlib.resources import path
from argparse import ArgumentParser, Namespace
from typing import Union

from jinja2 import Environment, FileSystemLoader
from markdown import Markdown
import yafg
from MarkdownHighlight.highlight import HighlightExtension
from markdown_checklist.extension import ChecklistExtension

from .configuration import Configuration
from .database import Database
from .builder import Builder
from .page import Page


def get_options() -> Namespace:
    parser = ArgumentParser(prog='pyssg',
                            description='''Static Site Generator that reads
                            Markdown files and creates HTML files.\nIf
                            [-c]onfig file is provided (or exists in default
                            location) all other options are ignored.\nFor
                            datetime formats see:
                            https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes''')
    parser.add_argument('-v', '--version',
                        action='store_true',
                        help='''print program version''')
    parser.add_argument('-c', '--config',
                        default='$XDG_CONFIG_HOME/pyssg/pyssgrc',
                        type=str,
                        help='''config file (path) to read from; defaults to
                        'pyssgrc' first, then
                        '$XDG_CONFIG_HOME/pyssg/pyssgrc' ''')
    parser.add_argument('-s', '--src',
                        default='src',
                        type=str,
                        help='''src directory; handmade files, templates and
                        metadata directory; defaults to 'src' ''')
    parser.add_argument('-d', '--dst',
                        default='dst',
                        type=str,
                        help='''dst directory; generated (and transfered html)
                        files; defaults to 'dst' ''')
    parser.add_argument('-t', '--plt',
                        default='plt',
                        type=str,
                        help='''plt directory; all template files; defaults to
                        'plt' ''')
    parser.add_argument('-u', '--url',
                        default='',
                        type=str,
                        help='''base url without trailing slash''')
    parser.add_argument('--static-url',
                        default='',
                        type=str,
                        help='''base static url without trailing slash''')
    parser.add_argument('--default-image-url',
                        default='',
                        type=str,
                        help='''default image url''')
    parser.add_argument('--title',
                        default='Blog',
                        type=str,
                        help='''general title for the website; defaults to
                        'Blog' ''')
    parser.add_argument('--date-format',
                        default='%a, %b %d, %Y @ %H:%M %Z',
                        type=str,
                        help='''date format used inside pages (for creation and
                        modification times, for example); defaults to '%%a, %%b
                        %%d, %%Y @ %%H:%%M %%Z' ('Tue, Mar 16, 2021 @ 02:46 UTC',
                        for example)''')
    parser.add_argument('--list-date-format',
                        default='%b %d',
                        type=str,
                        help='''date format used for page entries in a list;
                        defaults to '%%b %%d' ('Mar 16', for example)''')
    parser.add_argument('--list-sep-date-format',
                        default='%B %Y',
                        type=str,
                        help='''date format used for the separator between page
                        entries in a list; defaults to '%%B %%Y' ('March 2021',
                        for example)''')
    parser.add_argument('-i', '--init',
                        action='store_true',
                        help='''initializes the dir structure, templates,
                        as well as the 'src' and 'dst' directories''')
    parser.add_argument('-b', '--build',
                        action='store_true',
                        help='''generates all html files and passes over
                        existing (handmade) ones''')
    parser.add_argument('-f', '--force',
                        action='store_true',
                        help='''force building all pages and not only the
                        updated ones''')

    return parser.parse_args()


def main() -> None:
    opts: dict[str, Union[str, bool]] = vars(get_options())
    conf_path: str = opts['config']
    conf_path = os.path.expandvars(conf_path)


    config: Configuration = None
    if os.path.exists('pyssgrc'):
        config = Configuration('pyssgrc')
    else:
        config = Configuration(conf_path)

    config.read()
    config.fill_missing(opts)

    if opts['version']:
        print(f'pyssg v{config.version}')
        return

    if opts['init']:
        try:
            os.mkdir(config.src)
            os.makedirs(os.path.join(config.dst, 'tag'))
            os.mkdir(config.plt)
        except FileExistsError:
            pass

        # copy basic template files
        files: list[str] = ('index.html',
                            'page.html',
                            'tag.html',
                            'rss.xml',
                            'sitemap.xml')
        for f in files:
            plt_file: str = os.path.join(config.plt, f)
            with path('pyssg.plt', f) as p:
                if not os.path.exists(plt_file):
                    shutil.copy(p, plt_file)

        return

    if opts['build']:
        # start the db
        db: Database = Database(os.path.join(config.src, '.files'))
        db.read()

        # the autoescape option could be a security risk if used in a dynamic
        # website, as far as i can tell
        env: Environment = Environment(loader=FileSystemLoader(config.plt),
                                       autoescape=False,
                                       trim_blocks=True,
                                       lstrip_blocks=True)


        # md extensions
        exts: list = ['extra',
                      'meta',
                      'sane_lists',
                      'smarty',
                      'toc',
                      'wikilinks',
                      yafg.YafgExtension(stripTitle=True,
                                         figureClass="",
                                         figcaptionClass="",
                                         figureNumbering=False,
                                         figureNumberClass="number",
                                         figureNumberText="Figure"),
                      HighlightExtension,
                      ChecklistExtension()]
        md: Markdown = Markdown(extensions=exts,
                                output_format='html5')
        builder: Builder = Builder(config, env, db, md)
        builder.build()

        db.write()
        return



from datetime import datetime, timezone

from .configuration import Configuration


class Page:
    def __init__(self,
                 name: str,
                 ctime: float,
                 mtime: float,
                 html: str,
                 meta: dict,
                 config: Configuration):
        # initial data
        self.name: str = name
        self.ctimestamp: float = ctime
        self.mtimestamp: float = mtime
        self.content: str = html
        self.meta: dict = meta
        self.config: Configuration = config

        # data from self.meta
        self.title: str = ''
        self.author: str = ''
        self.cdatetime: datetime = None
        self.mdatetime: datetime = None
        self.summary: str = ''
        self.lang: str = 'en'
        self.tags: list[tuple[str]] = []

        # constructed
        self.url: str = ''
        self.image_url: str = ''
        self.cdate: str = ''
        self.cdate_list: str = ''
        self.cdate_list_sep: str = ''
        self.cdate_rss: str = ''
        self.cdate_sitemap: str = ''
        self.mdate: str = None
        self.mdate_list: str = None
        self.mdate_list_sep: str = None
        self.mdate_rss: str = ''
        self.mdate_sitemap: str = ''

        # later assigned references to next and previous pages
        self.next: Page = None
        self.previous: Page = None

        # also from self.meta, but for og metadata
        self.og: dict[str, str] = dict()


    def __lt__(self, other):
        return self.ctimestamp < other.ctimestamp


    # parses meta from self.meta, for og, it prioritizes,
    # the actual og meta
    def parse(self):
        # required meta elements
        self.title = self.meta['title'][0]
        self.author = self.meta['author'][0]
        self.summary = self.meta['summary'][0]
        self.lang = self.meta['lang'][0]

        # dates
        self.cdatetime = datetime.fromtimestamp(self.ctimestamp,
                                                 tz=timezone.utc)
        self.cdate = self.cdatetime.strftime(self.config.dformat)
        self.cdate_list = self.cdatetime.strftime(self.config.l_dformat)
        self.cdate_list_sep = self.cdatetime.strftime(self.config.lsep_dformat)
        self.cdate_rss = self.cdatetime.strftime(self.config.dformat_rss)
        self.cdate_sitemap = \
        self.cdatetime.strftime(self.config.dformat_sitemap)

        # only if file/page has been modified
        if self.mtimestamp != 0.0:
            self.mdatetime = datetime.fromtimestamp(self.mtimestamp,
                                                     tz=timezone.utc)
            self.mdate = self.mdatetime.strftime(self.config.dformat)
            self.mdate_list = self.mdatetime.strftime(self.config.l_dformat)
            self.mdate_list_sep = self.mdatetime.strftime(self.config.lsep_dformat)
            self.mdate_rss = self.mdatetime.strftime(self.config.dformat_rss)
            self.mdate_sitemap = \
            self.mdatetime.strftime(self.config.dformat_sitemap)

        # not always contains tags
        try:
            tags_only: list[str] = self.meta['tags']
            tags_only.sort()

            for t in tags_only:
                self.tags.append((t,
                                  f'{self.config.url}/tag/@{t}.html'))
        except KeyError: pass

        self.url = f'{self.config.url}/{self.name.replace(".md", ".html")}'

        try:
            self.image_url = \
            f'{self.config.static_url}/{self.meta["image_url"][0]}'
        except KeyError:
            self.image_url = \
            f'{self.config.static_url}/{self.config.default_image_url}'

        # if contains open graph elements
        try:
            # og_e = object graph entry
            for og_e in self.meta['og']:
                kv: str = og_e.split(',', 1)
                if len(kv) != 2:
                    raise Exception('invalid og syntax')

                k: str = kv[0].strip()
                v: str = kv[1].strip()

                self.og[k] = v
        except KeyError: pass



import os


# db class that works for both html and md files
class Database:
    def __init__(self, db_path: str):
        self.db_path: str = db_path
        self.e: dict[str, tuple[float, float, list[str]]] = dict()


    # updates the tags for a specific entry (file)
    def update_tags(self, file_name: str,
                    tags: list[str]) -> None:
        if file_name in self.e:
            cts, mts, _ = self.e[file_name]
            self.e[file_name] = (cts, mts, tags)


    # returns a bool that indicates if the entry
    # was (includes new entries) or wasn't updated
    # 0.0 means no mod
    def update(self, file_name: str,
               remove: str=None) -> bool:
        # initial default values
        f: str = file_name
        tags: list[str] = []
        if remove is not None:
            f = file_name.replace(remove, '')


        # get current time, needs actual file name
        time: float = os.stat(file_name).st_mtime

        # three cases, 1) entry didn't exist,
        # 2) entry hasn't been mod and,
        # 3) entry has been mod
        #1)
        if f not in self.e:
            self.e[f] = (time, 0.0, tags)
            return True

        old_time, old_mod_time, tags = self.e[f]

        # 2)
        if old_mod_time == 0.0:
            if time > old_time:
                self.e[f] = (old_time, time, tags)
                return True
        # 3)
        else:
            if time > old_mod_time:
                self.e[f] = (old_time, time, tags)
                return True

        return False


    def write(self) -> None:
        with open(self.db_path, 'w') as file:
            # write each k,v pair in dict to db file
            for k, v in self.e.items():
                t: str = None
                if len(v[2]) == 0:
                    t = '-'
                else:
                    t = ','.join(v[2])
                file.write(f'{k} {v[0]} {v[1]} {t}\n')


    def read(self) -> None:
        # only if the path exists and it is a file
        if os.path.exists(self.db_path) and os.path.isfile(self.db_path):
            # get all db file lines
            lines: list[str] = None
            with open(self.db_path, 'r') as file:
                lines = file.readlines()

            # parse each entry and populate accordingly
            l: list[str] = None
            # l=list of values in entry
            for line in lines:
                l = tuple(line.strip().split())
                if len(l) != 4:
                    raise Exception('db entry doesn\'t contain 4 elements')

                t: list[str] = None
                if l[3] == '-':
                    t = []
                else:
                    t = l[3].split(',')

                self.e[l[0]] = (float(l[1]), float(l[2]), t)



import os
from operator import itemgetter
from datetime import datetime
from markdown import Markdown

from .database import Database
from .configuration import Configuration
from .page import Page


# parser of md files, stores list of pages and tags
class MDParser:
    def __init__(self, src: str,
                 files: list[str],
                 config: Configuration,
                 db: Database,
                 md: Markdown):
        self.src: str = src
        self.files: list[str] = files

        self.config: Configuration = config
        self.db: Database = db
        self.md: Markdown = md

        self.all_pages: list[Page] = None
        self.updated_pages: list[Page] = None
        self.all_tags: list[tuple[str]] = None


    def parse(self) -> None:
        # initialize lists
        self.all_pages = []
        self.updated_pages = []
        self.all_tags = []
        all_tag_names: list[str] = []

        for f in self.files:
            src_file: str = os.path.join(self.src, f)
            # get flag if update is successful
            updated: bool = self.db.update(src_file, remove=f'{self.src}/')

            content: str = self.md.reset().convert(open(src_file).read())
            page: Page = Page(f,
                              self.db.e[f][0],
                              self.db.e[f][1],
                              content,
                              self.md.Meta,
                              self.config)
            page.parse()

            # keep a separated list for all and updated pages
            if updated:
                self.updated_pages.append(page)
            self.all_pages.append(page)

            # parse tags
            if page.tags is not None:
                # add its tag to corresponding db entry if existent
                self.db.update_tags(f, list(map(itemgetter(0), page.tags)))

                # update all_tags attribute
                for t in page.tags:
                    if t[0] not in list(map(itemgetter(0), self.all_tags)):
                        self.all_tags.append(t)

        # sort list of tags for consistency
        self.all_tags.sort(key=itemgetter(0))
        self.updated_pages.sort(reverse=True)
        self.all_pages.sort(reverse=True)

        pages_amount: int = len(self.all_pages)
        # note that prev and next are switched because of the reverse rodering
        # of all_pages
        for i, p in enumerate(self.all_pages):
            if i != 0:
                next_page: Page = self.all_pages[i - 1]
                p.next = next_page

            if i != pages_amount - 1:
                prev_page: Page = self.all_pages[i + 1]
                p.previous = prev_page



import os


def get_file_list(directory: str,
                  extensions: list[str],
                  exclude: list[str]=None) -> list[str]:
    out: list[str] = []
    for root, dirs, files in os.walk(directory):
        if exclude is not None:
            dirs[:] = [d for d in dirs if d not in exclude]

        for f in files:
            if f.endswith(tuple(extensions)):
                out.append(os.path.join(root, f).replace(directory, '')[1:])

    return out


def get_dir_structure(directory: str,
                      exclude: list[str]=None) -> list[str]:
    out: list[str] = []
    for root, dirs, files in os.walk(directory):
        if exclude is not None:
            dirs[:] = [d for d in dirs if d not in exclude]

        for d in dirs:
            if root in out:
                out.remove(root)
            out.append(os.path.join(root, d))

    return [o.replace(directory, '')[1:] for o in out]



import os
import shutil
from copy import deepcopy
from operator import itemgetter
from jinja2 import Environment, Template
from markdown import Markdown

from .configuration import Configuration
from .database import Database
from .parser import MDParser
from .page import Page
from .discovery import get_file_list, get_dir_structure


class Builder:
    def __init__(self, config: Configuration,
                 env: Environment,
                 db: Database,
                 md: Markdown):
        self.config: Configuration = config
        self.env: Environment = env
        self.db: Database = db
        self.md: Markdown = md

        self.dirs: list[str] = None
        self.md_files: list[str] = None
        self.html_files: list[str] = None

        self.all_pages: list[Page] = None
        self.updated_pages: list[Page] = None
        self.all_tags: list[str] = None
        self.common_vars: dict = None


    def build(self) -> None:
        self.dirs = get_dir_structure(self.config.src, ['templates'])
        self.md_files = get_file_list(self.config.src, ['.md'], ['templates'])
        self.html_files = get_file_list(self.config.src, ['.html'], ['templates'])

        self.__create_dir_structure()
        self.__copy_html_files()

        parser: MDParser = MDParser(self.config.src,
                                    self.md_files,
                                    self.config,
                                    self.db,
                                    self.md)
        parser.parse()

        # just so i don't have to pass these vars to all the functions
        self.all_pages = parser.all_pages
        self.updated_pages = parser.updated_pages
        self.all_tags = parser.all_tags

        # dict for the keyword args to pass to the template renderer
        self.common_vars = dict(config=self.config,
                                all_pages=self.all_pages,
                                all_tags=self.all_tags)

        self.__render_articles()
        self.__render_tags()
        self.__render_template('index.html', 'index.html', **self.common_vars)
        self.__render_template('rss.xml', 'rss.xml', **self.common_vars)
        self.__render_template('sitemap.xml', 'sitemap.xml', **self.common_vars)


    def __create_dir_structure(self) -> None:
        for d in self.dirs:
            # for the dir structure,
            # doesn't matter if the dir already exists
            try:
                os.makedirs(os.path.join(self.config.dst, d))
            except FileExistsError:
                pass


    def __copy_html_files(self) -> None:
        src_file: str = None
        dst_file: str = None

        for f in self.html_files:
            src_file = os.path.join(self.config.src, f)
            dst_file = os.path.join(self.config.dst, f)

            # only copy files if they have been modified (or are new)
            if self.db.update(src_file, remove=f'{self.config.src}/'):
                shutil.copy2(src_file, dst_file)


    def __render_articles(self) -> None:
        article_vars: dict = deepcopy(self.common_vars)
        # check if only updated should be created
        if self.config.force:
            for p in self.all_pages:
                article_vars['page'] = p
                self.__render_template("page.html",
                                       p.name.replace('.md','.html'),
                                       **article_vars)
        else:
            for p in self.updated_pages:
                article_vars['page'] = p
                self.__render_template("page.html",
                                       p.name.replace('.md','.html'),
                                       **article_vars)


    def __render_tags(self) -> None:
        tag_vars: dict = deepcopy(self.common_vars)
        for t in self.all_tags:
            # get a list of all pages that have current tag
            tag_pages: list[Page] = []
            for p in self.all_pages:
                if p.tags is not None and t[0] in list(map(itemgetter(0),
                                                           p.tags)):
                    tag_pages.append(p)

            tag_vars['tag'] = t
            tag_vars['tag_pages'] = tag_pages

            # build tag page
            self.__render_template('tag.html',
                                   f'tag/@{t[0]}.html',
                                   **tag_vars)

            # clean list of pages with current tag
            tag_pages = []


    def __render_template(self, template_name: str,
                          file_name: str,
                          **template_vars) -> None:
        template: Template = self.env.get_template(template_name)
        content: str = template.render(**template_vars)

        with open(os.path.join(self.config.dst, file_name), 'w') as f:
            f.write(content)
